#include "typedef.h"

#include <stdio.h>
#include <string.h>

#include "object.h"
#include "util.h"
#include "method.h"
#include "variable.h"
#include "class.h"
#include "edge.h"

Cvariable	*Cvariable::g_hash[HASHSIZE] = {0};

extern int g_zero_indirections;

/* static */ u4T
Cvariable::hash(Centity *entityP, int number)
{
	return((((int) entityP) + number) & (HASHSIZE-1));
}

Cvariable::Cvariable(Cmethod *parentP, int number, const u2T *nameP, u2T *descriptorP)
{
	Cclass	*classP;
	u2T		*P;
	int		parameters;

	m_hashP          = 0;
	m_parentP        = parentP;
	m_number         = number;
	m_nameP          = 0;
	if (nameP) {
		m_nameP      = unicodedup(nameP);
	} 
	m_access_flags   = 0;
	m_parameter      = 0;

	parameters = parentP->m_parameters;
	if (!(parentP->m_access_flags & 0x8)) {
		// Method is not static
		++parameters;
	}
	if (number < parameters) {
		m_parameter = 1;
	}

	add_created();

	m_descriptorP    = 0;
	if (descriptorP) {
		m_descriptorP = unicodedup(descriptorP);
		for (P = descriptorP; *P == '['; ++P);
		if (*P == 'L') {
			// Only create edges to classes or arrays of classes
			if (!g_zero_indirections) {
				P = descriptorP;
			}
			classP = Cclass::locate(P);
			new Cedge(this, classP, INSTANTIATES_EDGE, -1);
	}	}
}

Cvariable *
Cvariable::locate(Cmethod *parentP, int number, const u2T *nameP, u2T *descriptorP)
{
	unsigned int	index;
	Cvariable		**variablePP, *variableP;

	index         = hash(parentP, number);
	for (variablePP = g_hash + index; variableP = *variablePP; variablePP = &(variableP->m_hashP)) {
		if (variableP->m_parentP != parentP) {
			continue;
		}
		if (variableP->m_number != number) {
			continue;
		}
		return(variableP);
	}	
	variableP = new Cvariable(parentP, number, nameP, descriptorP);
	if (!variableP) {
		outofmemory();
	}
	*variablePP = variableP;
	return(variableP);
}

/* virtual */ objectE
Cvariable::type(void) const
{
	return(variableE);
}

/* virtual */ void
Cvariable::emit_id(void) const
{
	fprintf(stdout, "L%p", this);
}

const char *
Cvariable::get_type(void) const
{
	if (m_parameter) {
		return("P");
	}
	return("L");
}
		
/* virtual */ void
Cvariable::emit_instance(void) const
{
	fprintf(stdout, "$INSTANCE L%p %s\n", this, get_type());
}

void
Cvariable::emit_attributes(void) const
{
	emit_id();
	fputs(" {label=", stdout);
	if (m_nameP) {
		fputc('"', stdout);
		dumpUnicode(m_nameP);
		fputc('"', stdout);
	} else {
		fprintf(stdout, "%d", m_number);
	}
	fprintf(stdout, " no=%d", m_number);

	fputs(" file=.", stdout);

	if (m_descriptorP) {
		fputs(" descriptor=\"", stdout);
		dump_descriptor(m_descriptorP);
		fputc('"', stdout);
	}

	dump_access(m_access_flags);
	fputs("}\n", stdout);
}